#include "../std.h"

#include "../config/config_map.h"
#include "../transfer/rls.h"
#include "../misc/url_options.h"
#include "../misc/checksum.h"
#include "../misc/stringtoint.h"
#include "../misc/inttostring.h"
#include "../misc/time_utils.h"
#include "../misc/log_time.h"

#include "datapoint.h"

#ifdef HAVE_GLOBUS_RLS_CLIENT_H

#define globus_rls_free_result(err) globus_rls_client_error_info(err,NULL,NULL,0,GLOBUS_FALSE)


static globus_result_t globus_rls_client_lrc_attr_put(globus_rls_handle_t* h,char* key,globus_rls_attribute_t *attr,int overwrite) {
  globus_result_t err;
  int errcode;
  err = globus_rls_client_lrc_attr_add(h,key,attr);
  if(err != GLOBUS_SUCCESS) {
    err = globus_rls_client_error_info(err,&errcode,NULL,0,GLOBUS_TRUE);
    if((overwrite) && (errcode == GLOBUS_RLS_DBERROR)) { /* guess this can mean duplicate entry */
      globus_result_t err_;
      err_ = globus_rls_client_lrc_attr_remove(h,key,attr);
      globus_rls_free_result(err_);
      if(err_ != GLOBUS_SUCCESS) return err;
      return globus_rls_client_lrc_attr_put(h,key,attr,0);
    };
    if(errcode != GLOBUS_RLS_ATTR_NEXIST) return err;
    globus_rls_free_result(err);
    err = globus_rls_client_lrc_attr_create(h,attr->name,
                                               attr->objtype,attr->type);
    if(err != GLOBUS_SUCCESS) return err;
    err = globus_rls_client_lrc_attr_add(h,key,attr);
  };
  return err;
}

/*
void globus_list_join (globus_list_t * volatile * left,
                       globus_list_t * volatile * right) {
    globus_list_t * entry = GLOBUS_NULL;
    globus_list_t * last = left; 
    while(last) {
      entry=last; entry=globus_list_rest(entry);
    };
    if(last) {
      last->next=*right; *right=GLOBUS_NULL;
    } else {
      *left=*right; *right=GLOBUS_NULL;
    };
}

globus_result_t globus_rls_client_get_pfn(globus_rls_handle_t* h,const char* lfn,int offset,int limit,globus_list_t** pfns) {
  globus_list_t* lrcs = NULL;
  globus_list_t** pfns_;
  char errmsg[MAXERRMSG];
  globus_result_t err;
  int errcode;
  globus_rls_string2_t lrc_direct;
  lrc_direct.s1=(char*)(lfn);
  lrc_direct.s2=NULL; // for current connection


  int lrc_offset = 0;
  int lrc_limit = 100;
  if(lfn && lfn[0]) {
    err = globus_rls_client_rli_get_lrc(h,(char*)(lfn),&lrc_offset,lrc_limit,&lrcs);
  } else {
    err = globus_rls_client_rli_get_lrc_wc(h,"*",rls_pattern_unix,&lrc_offset,lrc_limit,&lrcs);
  };
  if(err != GLOBUS_SUCCESS) {
    globus_rls_client_error_info(err,&errcode,errmsg,MAXERRMSG,GLOBUS_TRUE);
    if(errcode != GLOBUS_RLS_INVSERVER) return err;
    globus_rls_free_result(err);
    // Probably that is LRC server only.
    globus_list_insert(lrcs,lrc_direct);
  };
  globus_list_t* p;
  err=GLOBUS_SUCCESS;
  // TODO: sort by lrc and cache connections
  for(p=lrcs;p;p=globus_list_rest(p)) {
    globus_rls_string2_t* str2 = (globus_rls_string2_t*)globus_list_first(p);
    char* lrc = str2->s2;
    globus_rls_handle_t* h_;
    if(err != GLOBUS_SUCCESS) {
      globus_rls_free_result(err); err=GLOBUS_SUCCESS;
    };
    if(lrc) {
      err = globus_rls_client_connect(lrc,&h_);
      if(err != GLOBUS_SUCCESS) { // hide error
        continue;
      };
    } else { h_=h; };
    if(lfn && lfn[0]) {
      err = globus_rls_client_lrc_get_pfn(h_,(char*)(lfn),&offset,limit,&pfns_);
    } else {
      err = globus_rls_client_lrc_get_pfn_wc(h_,"*",rls_pattern_unix,&offset,limit,&pfns_);
    };
    if(lrc) { globus_rls_client_close(h_); };
    if(err != GLOBUS_SUCCESS) { // hide error
      continue;
    };
    globus_list_join(&pfns,&pfns_);
  };
  globus_rls_client_free_list(lrcs);
  if(pfns && (err != GLOBUS_SUCCESS)) {
    globus_rls_free_result(err); err=GLOBUS_SUCCESS; 
  };
  return err;
}
*/
#endif


#ifdef HAVE_GLOBUS_RLS_CLIENT_H
typedef class meta_resolve_rls_t {
 public:
  DataPoint& url;
//  string& lfn;
  bool source;
  bool success;
  bool locations_empty;
  bool obtained_info;
  meta_resolve_rls_t(DataPoint& u,bool s):url(u),source(s),success(false),locations_empty(false),obtained_info(false) {
  };
};

bool DataPoint::meta_resolve_rls_callback(globus_rls_handle_t* h,const char* url,void* arg) {
  meta_resolve_rls_t* arg_ = (meta_resolve_rls_t*)arg;
  DataPoint& it = ((meta_resolve_rls_t*)arg)->url;
//  string& lfn = ((meta_resolve_rls_t*)arg)->lfn;
  string& lfn = it.rls_lfn;
  bool source = arg_->source;
  char errmsg[MAXERRMSG];
  globus_result_t err;
  int errcode;

  // Ask LRC if it contains file of interest
  globus_list_t* pfns_list = NULL;
  if(source) {
    err=globus_rls_client_lrc_get_pfn(h,(char*)(lfn.c_str()),0,0,&pfns_list);
  } else {
    err=globus_rls_client_lrc_get_pfn(h,(char*)("__storage_service__"),0,0,&pfns_list);
  };
  if(err != GLOBUS_SUCCESS) {
    globus_rls_client_error_info(err,&errcode,errmsg,MAXERRMSG,GLOBUS_FALSE);
    if(errcode == GLOBUS_RLS_INVSERVER) {
      return true;
    } else if(errcode == GLOBUS_RLS_LFN_NEXIST) {
      return true;
    } else { // do not know
      odlog(1)<<"Warning: can't get PFNs from server "<<url<<" : "<<errmsg<<endl;
      return true;
    };
  };
  if(!(arg_->success)) {
    arg_->success=true; // got something
    if(source) it.is_metaexisting=true;
    arg_->locations_empty=(it.locations.size() == 0);
  };
  if(arg_->locations_empty) {
    globus_list_t* list_p;
    for(list_p=pfns_list;list_p;list_p=globus_list_rest(list_p)) {
      globus_rls_string2_t* str2=
                       (globus_rls_string2_t*)globus_list_first(list_p);
      char* lfn = str2->s1;
      char* pfn = str2->s2;
      list<DataPoint::Location>::iterator loc =
        it.locations.insert(it.locations.end(),DataPoint::Location(url,pfn));
      loc->arg=(void*)1; // marker
      odlog(2)<<"Adding location: "<<url<<" - "<<pfn<<endl;
    };
  } else {
    list<Location>::iterator loc = it.locations.begin();
    for(;loc!=it.locations.end();++loc) {
      if(loc->arg != NULL) continue;
      globus_list_t* list_p;
      for(list_p=pfns_list;list_p;list_p=globus_list_rest(list_p)) {
        globus_rls_string2_t* str2=
                         (globus_rls_string2_t*)globus_list_first(list_p);
        char* lfn = str2->s1;
        char* pfn = str2->s2;
        // for RLS URLs are used instead of metanames
        if(strncmp(pfn,loc->meta.c_str(),loc->meta.length()) == 0) {
          odlog(2)<<"Adding location: "<<url<<" - "<<pfn<<endl;
          if(source) {
            loc->meta=url; loc->url=pfn;
          } else {
            loc->meta=url;
          };
          loc->arg=(void*)1; break;
        };
      };
    };
  };
  globus_rls_client_free_list(pfns_list);
  if(!arg_->obtained_info) {
    /* obtain metadata - assume it is same everywhere */
    globus_list_t* attr_list;
    err = globus_rls_client_lrc_attr_value_get(h,(char*)(lfn.c_str()),NULL,
                                  globus_rls_obj_lrc_lfn,&attr_list);
    if(err != GLOBUS_SUCCESS) {
      globus_rls_client_error_info(err,&errcode,errmsg,MAXERRMSG,GLOBUS_FALSE);
      if(errcode == GLOBUS_RLS_ATTR_NEXIST) {
        return true;
      };
      odlog(1)<<"Warning: Failed to obtain attributes from "<<url<<" : "<<errmsg<<endl;
      return true;
    };
    it.is_metaexisting=true; // even for destination
    for(globus_list_t* list_pa=attr_list;list_pa;
                                       list_pa=globus_list_rest(list_pa)) {
      unsigned long long int i;
      globus_rls_attribute_t* attr =
                          (globus_rls_attribute_t*)globus_list_first(list_pa);
      if(attr->type != globus_rls_attr_type_str) continue;
      odlog(2)<<"Attribute: "<<attr->name<<" - "<<attr->val.s<<endl;
      if(strcmp(attr->name,"filechecksum") == 0) {
        if(!it.meta_checksum_valid) 
          if(stringtoint(attr->val.s,i))it.meta_checksum(i);
      } else if(strcmp(attr->name,"size") == 0) {
        if(!it.meta_size_valid) 
          if(stringtoint(attr->val.s,i)) it.meta_size(i);
      } else if(strcmp(attr->name,"modifytime") == 0) {
        if(!it.meta_created_valid) 
          if(stringtoint(attr->val.s,i)) it.meta_created(i);
      } else if(strcmp(attr->name,"created") == 0) {
        if(!it.meta_created_valid) 
          if(stringtoint(attr->val.s,i)) it.meta_created(i);
      };
    };
    globus_rls_client_free_list(attr_list);
    arg_->obtained_info=true;
  };
  return true;
}
#endif

/* perform resolve operation, which can take long time */
bool DataPoint::meta_resolve_rls(bool source) {
#ifdef HAVE_GLOBUS_RLS_CLIENT_H
  is_resolved=false;
  if(source) {
    if(rls_lfn.length() == 0) {
      odlog(1)<<"Source must contain LFN"<<endl;
      return false;
    };
    list<string> rlis;
    list<string> lrcs;
    rlis.push_back(rls_url);
    lrcs.push_back(rls_url);
    meta_resolve_rls_t arg(*this,source);
    bool res = rls_find_lrcs(rlis,lrcs,true,false,&meta_resolve_rls_callback,(void*)(&arg));
    if(!arg.success) return false;
    // Remove unresolved locations
    list<Location>::iterator loc = locations.begin();
    for(;loc!=locations.end();) {
      if(loc->arg==NULL) {
        odlog(2)<<"Removing location: "<<loc->meta<<" - "<<loc->url<<endl;
        loc=locations.erase(loc);
      } else {
        odlog(2)<<"Using location: "<<loc->meta<<" - "<<loc->url<<endl;
        loc->arg=(void*)1; ++loc;
      };
    };
  } else { 
    if(rls_lfn.length() == 0) {
      odlog(1)<<"Destination must contain LFN"<<endl;
      return false;
    };
    list<string> rlis;
    list<string> lrcs;
    rlis.push_back(rls_url);
    lrcs.push_back(rls_url);
    if(locations.size() == 0) {
      odlog(1)<<"Warning: Locations are missing in destination RLS url - will use those registered with special name"<<endl;
    };
//    string storage_lfn("__storage_service__");
    meta_resolve_rls_t arg(*this,source);
    bool res = rls_find_lrcs(rlis,lrcs,true,false,&meta_resolve_rls_callback,(void*)(&arg));
    if(!arg.success) return false;
    if(locations.size() == 0) {
      odlog(1)<<"No locations found for destination"<<endl;
      return false;
    };
    // Make pfns
    list<Location>::iterator loc = locations.begin();
    list<string>::iterator lrc_p = lrcs.begin();
    for(;loc!=locations.end();) {
      if(strncasecmp(loc->url.c_str(),"se://",5) == 0) { loc->url+="?"; }
      else { loc->url+="/"; };
      loc->url+=rls_lfn;
      if(loc->arg!=NULL) {
        odlog(2)<<"Using location: "<<loc->meta<<" - "<<loc->url<<endl;
        ++loc;
      } else { // Use arbitrary lrc
        if(lrc_p == lrcs.end()) { // no LRC
          odlog(2)<<"Removing location: "<<loc->meta<<" - "<<loc->url<<endl;
          loc=locations.erase(loc);
        } else {
          loc->meta=*lrc_p; ++lrc_p; if(lrc_p==lrcs.end()) lrc_p=lrcs.begin();
          odlog(2)<<"Using location: "<<loc->meta<<" - "<<loc->url<<endl;
          loc->arg=(void*)1; ++loc;
        };
      };
    };
  };
  odlog(2)<<"meta_get_data: checksum: "<<meta_checksum()<<endl;
  odlog(2)<<"meta_get_data: size: "<<meta_size()<<endl;
  odlog(2)<<"meta_get_data: created: "<<meta_created()<<endl;
  location=locations.begin();
  is_resolved=true;
  return true;
#else
  return false;
#endif
}




/*
 



  if(!source) { * destination can't derive locations - they must be given ? *
    if(locations.size() == 0) {
      odlog(1)<<"Warning: Locations are missing in destination RLS url - will look for special name"<<endl;





  list<string> rlis;
  list<string> lrcs;
  rlis.push_back(rls_url);
  lrcs.push_back(rls_url);
  list_files_rls_t arg(files,*this,resolve);
  bool res = rls_find_lrcs(rlis,lrcs,true,false,&meta_resolve_rls_callback,(void*)(&arg));
  return arg.success;

















  is_resolved=false;
  if(!source) { * destination can't derive locations - 
                       they must be given *
    if(locations.size() == 0) {
      odlog(1)<<"Warning: Locations are missing in destination RLS url - will look for special name"<<endl;
      // go recursively and look for special name
      // code is ugly, but should work
      string rls_lfn_bak = rls_lfn;
      bool meta_checksum_valid_bak = meta_checksum_valid;
      unsigned long long meta_checksum_bak = meta_checksum_;
      bool meta_size_valid_bak = meta_size_valid;
      unsigned long long meta_size_bak = meta_size_;
      bool meta_created_valid_bak = meta_created_valid;
      unsigned long long meta_created_bak = meta_created_;

      rls_lfn="__storage_service__";
      bool res = meta_resolve_rls(true);
      
      meta_checksum_valid=meta_checksum_valid_bak;
      meta_checksum_=meta_checksum_bak;
      meta_size_valid=meta_size_valid_bak;
      meta_size_=meta_size_bak;
      meta_created_valid=meta_created_valid_bak;
      meta_created_=meta_created_bak;
      rls_lfn=rls_lfn_bak;

      // It is ok to have res == false. locattions.size() is more important
      if(locations.size() == 0) {
        odlog(1)<<"No locations found for destination"<<endl;
        return false;
      };
      // add lfn to locations
      for(list<Location>::iterator loc = locations.begin();
                                              loc!=locations.end();++loc) {
        if(strncasecmp(loc->url.c_str(),"se://",5) == 0) {
          loc->url+="?"; 
        } else {
          loc->url+="/"; 
        };
        loc->url+=rls_lfn;
        odlog(2)<<"location: "<<loc->meta<<" - "<<loc->url<<endl;
      };
      // continue for attributes
    };
  };
  is_metaexisting=false;



  globus_rls_handle_t* h = NULL;
  globus_list_t* pfns_list = NULL;
  list<string> lrcs; 
//  globus_bool_t free_lrcs = GLOBUS_FALSE;
  char errmsg[MAXERRMSG];
  globus_result_t err;
  int errcode;
//  globus_rls_string2_t lrc_direct;

//  lrc_direct.s1=(char*)(rls_lfn.c_str());
//  lrc_direct.s2=(char*)(rls_url.c_str());

  err = globus_rls_client_connect((char*)(rls_url.c_str()),&h);
  if(err != GLOBUS_SUCCESS) {
    globus_rls_client_error_info(err,&errcode,errmsg,MAXERRMSG,GLOBUS_FALSE);
    odlog(1)<<"Failed to connect to RLS server: "<<errmsg<<endl;
    return false;
  };

  // assume server is rli
  int lrc_offset = 0;
  globus_list_t* lrcs_ = NULL;
  err = globus_rls_client_rli_get_lrc(h,(char*)(rls_lfn.c_str()),
                      &lrc_offset,100,&lrcs_);
  if(err != GLOBUS_SUCCESS) {
    globus_rls_client_error_info(err,&errcode,errmsg,MAXERRMSG,GLOBUS_FALSE);
    if(errcode == GLOBUS_RLS_INVSERVER) {
      lrcs.push_back(rls_url);
//      globus_list_insert(&lrcs,&lrc_direct);
    } else if(errcode == GLOBUS_RLS_LFN_NEXIST) {
      err = globus_rls_client_rli_sender_list(h,&lrcs_);
      if(err != GLOBUS_SUCCESS) {
        globus_rls_client_error_info(err,&errcode,errmsg,MAXERRMSG,GLOBUS_FALSE);
        globus_rls_client_close(h);
        odlog(1)<<"Failed to retrieve list of LRCs: "<<errmsg<<endl;
        return false;
      };
      globus_list_t* list_p;
      for(list_p=lrcs_;list_p;list_p=globus_list_rest(list_p)) {
        globus_rls_sender_info_t* u=
                      (globus_rls_sender_info_t*)globus_list_first(list_p);
        lrcs.push_back(string(u->url));
      };
      globus_rls_client_free_list(lrcs_);
    } else {
      globus_rls_client_close(h);
      odlog(1)<<"Failed to retrieve list of LRCs: "<<errmsg<<endl;
      return false;
    };
  } else {
    globus_list_t* list_p;
    for(list_p=lrcs_;list_p;list_p=globus_list_rest(list_p)) {
      globus_rls_string2_t* str2=
                       (globus_rls_string2_t*)globus_list_first(list_p);
      lrcs.push_back(string(str2->s2));
    };
 //   free_lrcs=GLOBUS_TRUE;
    globus_rls_client_free_list(lrcs_);
  }
  globus_rls_client_close(h); h=NULL;
  list<string>::iterator list_p;
  bool obtained_info = false;
  for(list_p=lrcs.begin();list_p!=lrcs.end();++list_p) {
    const char* lrc = list_p->c_str();
    odlog(2)<<"Contacting LRC server: "<<lrc<<endl;
    err = globus_rls_client_connect((char*)lrc,&h);
    if(err != GLOBUS_SUCCESS) {
      globus_rls_client_error_info(err,&errcode,errmsg,MAXERRMSG,GLOBUS_FALSE);
      odlog(1)<<"Warning: Failed to connect to RLS server "<<lrc<<": "<<errmsg<<endl;
      continue;
    };
    err=globus_rls_client_lrc_get_pfn(h,(char*)(rls_lfn.c_str()),0,0,&pfns_list);
    if(err != GLOBUS_SUCCESS) {
      globus_rls_client_error_info(err,&errcode,errmsg,MAXERRMSG,GLOBUS_FALSE);
      if(errcode == GLOBUS_RLS_LFN_NEXIST) {
        globus_rls_client_close(h);
        if(source) {
          odlog(1)<<"Warning: Requested LFN (for source) is missing at "<<lrc<<endl;
          continue;
        };
        * ok for destination *
*
        is_resolved=true;
        return true;
*
        continue;
      };
      globus_rls_client_close(h);
      odlog(1)<<"Warning: Failed to obtain PFNs from "<<lrc<<" : "<<errmsg<<endl;
      continue;
    };
    is_metaexisting=true;
    if(source) {  * TODO: remove existing pfns from locations for !source *
      if(locations.size() == 0) {
        globus_list_t* list_p;
        for(list_p=pfns_list;list_p;list_p=globus_list_rest(list_p)) {
          globus_rls_string2_t* str2=
                         (globus_rls_string2_t*)globus_list_first(list_p);
          char* lfn = str2->s1;
          char* pfn = str2->s2;
          list<DataPoint::Location>::iterator loc =
              locations.insert(locations.end(),DataPoint::Location(lrc,pfn));
          loc->arg=(void*)1; // marker
          odlog(2)<<"Adding location: "<<lrc<<" - "<<pfn<<endl;
        }; 
      }
      else {
        list<Location>::iterator loc = locations.begin();
        for(;loc!=locations.end();) {
          globus_list_t* list_p;
          for(list_p=pfns_list;list_p;list_p=globus_list_rest(list_p)) {
            globus_rls_string2_t* str2=
                             (globus_rls_string2_t*)globus_list_first(list_p);
            char* lfn = str2->s1;
            char* pfn = str2->s2;
            // for RLS URLs are used instead of metanames
            if(strncmp(pfn,loc->meta.c_str(),loc->meta.length()) == 0) {
              loc->meta=lrc; loc->arg=(void*)1; break;
            };
          };
          ++loc;
        };
      };
    } // source
    else {
      list<Location>::iterator loc = locations.begin();
      for(;loc!=locations.end();) {
        loc->meta=lrc; loc->arg=(void*)1; ++loc;
      };
    };
    globus_rls_client_free_list(pfns_list);
    * obtain metadata - assume it is same everywhere *
    globus_list_t* attr_list;
    err = globus_rls_client_lrc_attr_value_get(h,(char*)(rls_lfn.c_str()),NULL,
                                  globus_rls_obj_lrc_lfn,&attr_list);
    if(err != GLOBUS_SUCCESS) {
      globus_rls_client_error_info(err,&errcode,errmsg,MAXERRMSG,GLOBUS_FALSE);
      if(errcode == GLOBUS_RLS_ATTR_NEXIST) { 
        globus_rls_client_close(h);
*
        is_resolved=true;
        return true;
*
        continue;
      };
      globus_rls_client_close(h);
      odlog(1)<<"Warning: Failed to obtain attributes from "<<lrc<<" : "<<errmsg<<endl;
      continue;
    };
    globus_rls_client_close(h);
    for(globus_list_t* list_pa=attr_list;list_pa;
                                       list_pa=globus_list_rest(list_pa)) {
      unsigned long long int i;
      globus_rls_attribute_t* attr = 
                          (globus_rls_attribute_t*)globus_list_first(list_pa);
      if(attr->type != globus_rls_attr_type_str) continue;
      odlog(2)<<"Attribute: "<<attr->name<<" - "<<attr->val.s<<endl;
      if(strcmp(attr->name,"filechecksum") == 0) {
        if(!meta_checksum_valid) if(stringtoint(attr->val.s,i))meta_checksum(i);
      } else if(strcmp(attr->name,"size") == 0) {
        if(!meta_size_valid) if(stringtoint(attr->val.s,i)) meta_size(i);
      } else if(strcmp(attr->name,"modifytime") == 0) {
        if(!meta_created_valid) if(stringtoint(attr->val.s,i)) meta_created(i);
      } else if(strcmp(attr->name,"created") == 0) {
        if(!meta_created_valid) if(stringtoint(attr->val.s,i)) meta_created(i);
      };
    };
    globus_rls_client_free_list(attr_list);
    obtained_info=true;
  };
  {
    list<Location>::iterator loc = locations.begin();
    for(;loc!=locations.end();) {
      if(loc->arg==NULL) { 
        odlog(2)<<"Removing location: "<<loc->meta<<" - "<<loc->url<<endl;
        loc=locations.erase(loc);
      } else { loc->arg=(void*)1; ++loc; };
    };
  };
  // if(free_lrcs) { globus_rls_client_free_list(lrcs); }
  // else { globus_list_free(lrcs); };
  if(!is_metaexisting) {
    if(!source) {
      * ok for destination *
      is_resolved=true;
      return true;
    };
    odlog(1)<<"Warning: Requested LFN (for source) is missing at all LRCs"<<endl;
    return false;
  };
  odlog(2)<<"meta_get_data: checksum: "<<meta_checksum()<<endl;
  odlog(2)<<"meta_get_data: size: "<<meta_size()<<endl;
  odlog(2)<<"meta_get_data: created: "<<meta_created()<<endl;
  location=locations.begin();
  is_resolved=true;
  return true;
#else
  return false;
#endif
}
*/

bool DataPoint::meta_preregister_rls(bool replication,bool force) {
#ifdef HAVE_GLOBUS_RLS_CLIENT_H
  if(replication) { /* replicating inside same lfn */
    if(!is_metaexisting) { /* for replication it must be there */
      odlog(0)<<"LFN is missing in RLS (needed for replication)"<<endl;
      return false;
    };
    return true;
  };
  if(is_metaexisting) { /* algorithm require this to be new file */
    if(!force) {
      odlog(0)<<"LFN already exists in replica"<<endl;
      return false;
    };
  };
  /* RLS does not support LFN only in database - hence doing nothing here */
  return true;
#else
  return false;
#endif
}

bool DataPoint::meta_postregister_rls(bool replication,bool failure) {
#ifdef HAVE_GLOBUS_RLS_CLIENT_H
  globus_rls_handle_t* h;
  char errmsg[MAXERRMSG];
  globus_result_t err;
  int errcode;

  err = globus_rls_client_connect((char*)(rls_url.c_str()),&h);
  if(err != GLOBUS_SUCCESS) {
    globus_rls_client_error_info(err,NULL,errmsg,MAXERRMSG,GLOBUS_FALSE);
    odlog(1)<<"Failed to connect to RLS server: "<<errmsg<<endl;
    return false;
  };
  // assume that is RLI and try to resolve for special/any name 


  if(!replication) {
    if((err = globus_rls_client_lrc_create(h,(char*)(rls_lfn.c_str()),
                  (char*)(location->url.c_str()))) != GLOBUS_SUCCESS) {
      err=globus_rls_client_error_info(err,&errcode,NULL,0,GLOBUS_TRUE);
      if(errcode == GLOBUS_RLS_LFN_EXIST) {
         globus_rls_free_result(err);
         err = globus_rls_client_lrc_add(h,(char*)(rls_lfn.c_str()),
                                           (char*)(location->url.c_str()));
      };
    };
  }
  else {
    err = globus_rls_client_lrc_add(h,(char*)(rls_lfn.c_str()),
                                           (char*)(location->url.c_str()));
  };
  if(err != GLOBUS_SUCCESS) {
    globus_rls_client_error_info(err,&errcode,errmsg,MAXERRMSG,GLOBUS_FALSE);
    if(errcode != GLOBUS_RLS_MAPPING_EXIST) {
      odlog(1)<<"Failed to create/add LFN-PFN mapping: "<<errmsg<<endl;
      globus_rls_client_close(h);
      return false;
    };
  };
  globus_rls_attribute_t attr;
  string attr_val;
  attr.objtype=globus_rls_obj_lrc_lfn;
  attr.type=globus_rls_attr_type_str;
  attr.name="filetype";
  attr.val.s="file";
  err=globus_rls_client_lrc_attr_put(h,(char*)(rls_lfn.c_str()),&attr,0);
  if(err != GLOBUS_SUCCESS) {
    globus_rls_client_error_info(err,&errcode,errmsg,MAXERRMSG,GLOBUS_FALSE);
    if(errcode != GLOBUS_RLS_ATTR_EXIST) {
      odlog(1)<<"Warning: failed to add attribute to RLS: "<<errmsg<<endl;
    };
  };
  if(meta_size_valid) {
    attr.name="size";
    attr_val=inttostring(meta_size_);
    attr.val.s=(char*)(attr_val.c_str());
    err=globus_rls_client_lrc_attr_put(h,(char*)(rls_lfn.c_str()),&attr,0);
    if(err != GLOBUS_SUCCESS) {
      globus_rls_client_error_info(err,&errcode,errmsg,MAXERRMSG,GLOBUS_FALSE);
      if(errcode != GLOBUS_RLS_ATTR_EXIST) {
        odlog(1)<<"Warning: failed to add attribute to RLS: "<<errmsg<<endl;
      };
    };
  };
  if(meta_checksum_valid) {
    attr.name="filechecksum";
    attr_val=inttostring(meta_checksum_);
    attr.val.s=(char*)(attr_val.c_str());
    err=globus_rls_client_lrc_attr_put(h,(char*)(rls_lfn.c_str()),&attr,0);
    if(err != GLOBUS_SUCCESS) {
      globus_rls_client_error_info(err,&errcode,errmsg,MAXERRMSG,GLOBUS_FALSE);
      if(errcode != GLOBUS_RLS_ATTR_EXIST) {
        odlog(1)<<"Warning: failed to add attribute to RLS: "<<errmsg<<endl;
      };
    };
  };
  if(meta_created_valid) {
    attr.name="modifytime";
    attr_val=inttostring(meta_created_);
    attr.val.s=(char*)(attr_val.c_str());
    err=globus_rls_client_lrc_attr_put(h,(char*)(rls_lfn.c_str()),&attr,0);
    if(err != GLOBUS_SUCCESS) {
      globus_rls_client_error_info(err,&errcode,errmsg,MAXERRMSG,GLOBUS_FALSE);
      if(errcode != GLOBUS_RLS_ATTR_EXIST) {
        odlog(1)<<"Warning: failed to add attribute to RLS: "<<errmsg<<endl;
      };
    };
  };
  globus_rls_client_close(h);
  return true;
#else
  return false;
#endif
}

bool DataPoint::meta_preunregister_rls(bool replication) {
#ifdef HAVE_GLOBUS_RLS_CLIENT_H
  if(replication) return true;
  return true;
#else
  return false;
#endif
}

bool DataPoint::meta_unregister_rls(bool all) {
#ifdef HAVE_GLOBUS_RLS_CLIENT_H
  if(!all) {
    if(location == locations.end()) { 
      odlog(0)<<"Location is missing"<<endl;
      return false;
    };
    if(strncasecmp(location->url.c_str(),"se://",5) == 0) {
      odlog(2)<<"SE location will be unregistered automatically"<<endl;
      return true;
    };
  };
  globus_rls_handle_t* h;
  char errmsg[MAXERRMSG];
  globus_result_t err;
  int errcode;
  globus_list_t* pfns_list;

  err = globus_rls_client_connect((char*)(rls_url.c_str()),&h);
  if(err != GLOBUS_SUCCESS) {
    globus_rls_client_error_info(err,NULL,errmsg,MAXERRMSG,GLOBUS_FALSE);
    odlog(1)<<"Failed to connect to RLS server: "<<errmsg<<endl;
    return false;
  };
  // first find all LRC servers storing required information
  globus_list_t* lrcs = NULL;
  globus_rls_string2_t lrc_direct;
  globus_bool_t free_lrcs = GLOBUS_FALSE;
  lrc_direct.s1=(char*)(rls_lfn.c_str());
  lrc_direct.s2=NULL; // for current connection
  int lrc_offset = 0;
  int lrc_limit = 0;
  err = globus_rls_client_rli_get_lrc(h,(char*)(rls_lfn.c_str()),&lrc_offset,lrc_limit,&lrcs);
  if(err != GLOBUS_SUCCESS) {
    globus_rls_client_error_info(err,&errcode,errmsg,MAXERRMSG,GLOBUS_FALSE);
    if(errcode != GLOBUS_RLS_INVSERVER) {
      odlog(1)<<"Failed to retrieve LFN/LRC: "<<errmsg<<endl;
      return false;
    };
    // Probably that is LRC server only.
    globus_list_insert(&lrcs,&lrc_direct);
  } else { free_lrcs=GLOBUS_TRUE; };
  globus_list_t* p;
  err=GLOBUS_SUCCESS;
  // TODO: sort by lrc and cache connections
  bool failure = false;
  for(p=lrcs;p;p=globus_list_rest(p)) {
    globus_rls_string2_t* str2 = (globus_rls_string2_t*)globus_list_first(p);
    char* lrc = str2->s2;
    globus_rls_handle_t* h_;
    if(lrc) {
      err = globus_rls_client_connect(lrc,&h_);
      if(err != GLOBUS_SUCCESS) {
        globus_rls_client_error_info(err,&errcode,errmsg,MAXERRMSG,GLOBUS_FALSE);
        odlog(1)<<"Warning: Failed to connect to LRC at "<<lrc<<" : "<<errmsg<<endl;
        failure=true; continue;
      };
    } else { h_=h; }; // This server is already connected
    if(all) {
      err = globus_rls_client_lrc_get_pfn(h_,(char*)(rls_lfn.c_str()),&lrc_offset,lrc_limit,&pfns_list);
      if(err != GLOBUS_SUCCESS) {
        globus_rls_client_error_info(err,&errcode,errmsg,MAXERRMSG,GLOBUS_FALSE);
        odlog(1)<<"Warning: Failed to retrieve LFN/PFNs from "<<lrc<<" : "<<errmsg<<endl;
        if(lrc) globus_rls_client_close(h_);
        failure=true; continue;
      };
      globus_list_t* list_p;
      for(list_p=pfns_list;list_p;list_p=globus_list_rest(list_p)) {
        globus_rls_string2_t* str2=
                         (globus_rls_string2_t*)globus_list_first(list_p);
        char* pfn = str2->s1;
        if(strncasecmp(pfn,"se://",5) == 0) {
          odlog(2)<<"SE location will be unregistered automatically"<<endl;
        } else {
          err = globus_rls_client_lrc_delete(h_,(char*)(rls_lfn.c_str()),pfn);
          if(err != GLOBUS_SUCCESS) {
            globus_rls_client_error_info(err,&errcode,errmsg,MAXERRMSG,GLOBUS_FALSE);
            if((errcode != GLOBUS_RLS_MAPPING_NEXIST) && (errcode != GLOBUS_RLS_LFN_NEXIST) && (errcode != GLOBUS_RLS_PFN_NEXIST)) {
              odlog(1)<<"Warning: Failed to delete LFN/PFN from "<<lrc<<" : "<<errmsg<<endl;
              failure=true; continue;
            };
          };
        };
      };
      globus_rls_client_free_list(pfns_list);
    } else { // ! all
      err = globus_rls_client_lrc_delete(h_,(char*)(rls_lfn.c_str()),(char*)(location->url.c_str()));
      if(err != GLOBUS_SUCCESS) {
        globus_rls_client_error_info(err,&errcode,errmsg,MAXERRMSG,GLOBUS_FALSE);
        if((errcode != GLOBUS_RLS_MAPPING_NEXIST) && 
           (errcode != GLOBUS_RLS_LFN_NEXIST) &&
           (errcode != GLOBUS_RLS_PFN_NEXIST)) {
          odlog(1)<<"Warning: Failed to delete LFN/PFN from "<<lrc<<" : "<<errmsg<<endl;
          if(lrc) globus_rls_client_close(h_);
          failure=true; continue;
        };
      };
    };
    if(lrc) globus_rls_client_close(h_);
  };
  globus_rls_client_close(h);
  if(free_lrcs) { globus_rls_client_free_list(lrcs); }
  else { globus_list_free(lrcs); };
  return !failure;
#else
  return false;
#endif
}

/* rls://[url[|url[...]]@]server/lfn */
bool DataPoint::process_meta_url_rls(const string &url,string &cat_url,string &filename,list<DataPoint::Location> &locations) {
  if(!strncasecmp(url.c_str(),"rls://",6)) {
    cat_url.resize(0); locations.clear(); filename.resize(0);
    string url_(url.c_str());
    /* find out if it contains locations */
    int loc_start=6;
    int loc_end=url_.find('@',loc_start);
    string urls("");
    if(loc_end!=string::npos) {
      urls=url_.substr(loc_start,loc_end-loc_start);
      url_.erase(loc_start,loc_end-loc_start+1);
    };
    /* get lfn */
    int server_start=6;
    int server_end=url_.find('/',loc_start);
    if(server_end==string::npos) {
//      odlog(1)<<"LFN is missing in url "<<url_<<endl;
//      return false;
      filename=""; cat_url=url_;
    } else {
      filename=url_.substr(server_end+1);
      cat_url=url_.substr(0,server_end);
    };
    odlog(2)<<"LFN: "<<filename<<endl;
    odlog(2)<<"RLS server: "<<cat_url<<endl;
    odlog(2)<<"Location urls: "<<urls<<endl;
    int n=0;
    for(int nn=0;n<urls.length();) {
      nn=urls.find('|',n); if(nn == string::npos) nn=urls.length();
      if(n == nn) { n++; continue; };
      string loc(urls.c_str()+n,nn-n);
      locations.push_back(DataPoint::Location(loc.c_str(),loc.c_str()));
      n=nn+1;
    };
    return true;
  };
  return false;
};

bool get_attributes(globus_rls_handle_t* h,DataPoint::FileInfo &f) {
  globus_list_t* attr_list;
  char errmsg[MAXERRMSG];
  globus_result_t err;
  int errcode;
  err = globus_rls_client_lrc_attr_value_get(h,(char*)(f.name.c_str()),NULL,
                                  globus_rls_obj_lrc_lfn,&attr_list);
 if(err != GLOBUS_SUCCESS) {
   globus_rls_client_error_info(err,&errcode,errmsg,MAXERRMSG,GLOBUS_FALSE);
   if(errcode != GLOBUS_RLS_ATTR_NEXIST) {
     odlog(1)<<"Warning: Failed to retrieve attributes: "<<errmsg<<endl;
     return false;
   };
   return true;
 };
 for(globus_list_t* pa=attr_list;pa;pa=globus_list_rest(pa)) {
   globus_rls_attribute_t* attr =
                        (globus_rls_attribute_t*)globus_list_first(pa);
   if(attr->type != globus_rls_attr_type_str) continue;
   odlog(2)<<"Attribute: "<<attr->name<<" - "<<attr->val.s<<endl;
   if(strcmp(attr->name,"filechecksum") == 0) {
     if(stringtoint(attr->val.s,f.checksum)) f.checksum_available=true;
   } else if(strcmp(attr->name,"size") == 0) {
     if(stringtoint(attr->val.s,f.size)) f.size_available=true;
   } else if(strcmp(attr->name,"modifytime") == 0) {
     if(stringtotime(f.created,string(attr->val.s))==0)f.created_available=true;
   } else if(strcmp(attr->name,"created") == 0) {
     if(stringtotime(f.created,string(attr->val.s))==0)f.created_available=true;
   };
 };
 globus_rls_client_free_list(attr_list);
 return true;
}

#ifdef HAVE_GLOBUS_RLS_CLIENT_H
typedef class list_files_rls_t {
 public:
  list<DataPoint::FileInfo> &files;
  DataPoint& url;
  bool success;
  bool resolve;
  list_files_rls_t(list<DataPoint::FileInfo> &f,DataPoint& u,bool r):files(f),url(u),resolve(r),success(false) {
  };
};

bool DataPoint::list_files_rls_callback(globus_rls_handle_t* h,const char* url,void* arg) {
  list_files_rls_t* arg_ = (list_files_rls_t*)arg;
  DataPoint& it = ((list_files_rls_t*)arg)->url;
  list<DataPoint::FileInfo>& files = ((list_files_rls_t*)arg)->files; 
  int lrc_offset = 0;
  globus_result_t err;
  int errcode;
  char errmsg[MAXERRMSG];
  globus_list_t* pfns = NULL;
  if(it.rls_lfn.length()) {
    err = globus_rls_client_lrc_get_pfn(h,(char*)(it.rls_lfn.c_str()),
                          &lrc_offset,1000,&pfns);
  } else {
    err = globus_rls_client_lrc_get_pfn_wc(h,"*",rls_pattern_unix,
                          &lrc_offset,1000,&pfns);
  };
  if(err != GLOBUS_SUCCESS) {
    globus_rls_client_error_info(err,&errcode,errmsg,MAXERRMSG,GLOBUS_FALSE);
    if(errcode == GLOBUS_RLS_LFN_NEXIST) {
      odlog(2)<<"No LFNs found in "<<url<<endl;
      ((list_files_rls_t*)arg)->success=true;
      return true;
    };
    odlog(1)<<"Failed to retrieve list of LFNs/PFNs from "<<url<<" : "<<errmsg<<endl;
    return true;
  };
  ((list_files_rls_t*)arg)->success=true;
  globus_list_t* p;
  for(p=pfns;p;p=globus_list_rest(p)) {
    globus_rls_string2_t* str2=
                   (globus_rls_string2_t*)globus_list_first(p);
    char* lfn = str2->s1;
    char* pfn = str2->s2;
    odlog(2)<<"lfn: "<<lfn<<" - "<<" pfn: "<<pfn<<endl;
    list<DataPoint::FileInfo>::iterator f;
    for(f=files.begin();f != files.end();++f) if(f->name == lfn) break;
    if(f == files.end()) f=files.insert(files.end(),DataPoint::FileInfo(lfn));
    if(((list_files_rls_t*)arg)->resolve) get_attributes(h,*f);
    f->urls.push_back(string(pfn));
  };
  globus_rls_client_free_list(pfns);
  return true;
}
#endif

bool DataPoint::list_files_rls(list<DataPoint::FileInfo> &files,bool resolve) {
#ifdef HAVE_GLOBUS_RLS_CLIENT_H
  list<string> rlis;
  list<string> lrcs;
  rlis.push_back(rls_url);
  lrcs.push_back(rls_url);
  list_files_rls_t arg(files,*this,resolve);
  bool res = rls_find_lrcs(rlis,lrcs,true,false,&list_files_rls_callback,(void*)(&arg));
  return arg.success;
#else
  return false;
#endif
}


